package proj2

// You MUST NOT change what you import.  If you add ANY additional
// imports it will break the autograder, and we will be Very Upset.

import (
	// You neet to add with
	// go get github.com/nweaver/cs161-p2/userlib
	"github.com/nweaver/cs161-p2/userlib"

	// Life is much easier with json:  You are
	// going to want to use this so you can easily
	// turn complex structures into strings etc...
	"encoding/json"

	// Likewise useful for debugging etc
	"encoding/hex"

	// UUIDs are generated right based on the crypto RNG
	// so lets make life easier and use those too...
	//
	// You need to add with "go get github.com/google/uuid"
	"github.com/google/uuid"

	// Useful for debug messages, or string manipulation for datastore keys
	"strings"

	// Want to import errors
	"errors"

	// optional
	// strconv

	// if you are looking for fmt, we don't give you fmt, but you can use userlib.DebugMsg
	// see someUsefulThings() below
)

// This serves two purposes: It shows you some useful primitives and
// it suppresses warnings for items not being imported
func someUsefulThings() {
	// Creates a random UUID
	f := uuid.New()
	userlib.DebugMsg("UUID as string:%v", f.String())

	// Example of writing over a byte of f
	f[0] = 10
	userlib.DebugMsg("UUID as string:%v", f.String())

	// takes a sequence of bytes and renders as hex
	h := hex.EncodeToString([]byte("fubar"))
	userlib.DebugMsg("The hex: %v", h)

	// Marshals data into a JSON representation
	// Will actually work with go structures as well
	d, _ := json.Marshal(f)
	userlib.DebugMsg("The json data: %v", string(d))
	var g uuid.UUID
	json.Unmarshal(d, &g)
	userlib.DebugMsg("Unmashaled data %v", g.String())

	// This creates an error type
	userlib.DebugMsg("Creation of error %v", errors.New(strings.ToTitle("This is an error")))

	// And a random RSA key.  In this case, ignoring the error
	// return value
	var pk userlib.PKEEncKey
	var sk userlib.PKEDecKey
	pk, sk, _ = userlib.PKEKeyGen()
	userlib.DebugMsg("Key is %v, %v", pk, sk)
}

// Helper function: Takes the first 16 bytes and
// converts it into the UUID type
func bytesToUUID(data []byte) (ret uuid.UUID) {
	for x := range ret {
		ret[x] = data[x]
	}
	return
}

//the structure definition of a file key for one whole file
type FileKeys struct {
	Filename     string
	SymmetricKey []byte
	Signkey      []byte
	FileUUID     uuid.UUID
}

// The structure definition for a file record
type File struct {
	Filename     string
	FileUUID     uuid.UUID
	SymmetricKey []byte
	Signkey      []byte
	FileSlice    int
}

// The structure definition for a user record
type User struct {
	Username string
	Pubk     userlib.PKEEncKey
	Privk    userlib.PKEDecKey
	Signk    userlib.DSSignKey
	Vrfyk    userlib.DSVerifyKey
	Filemap  map[string](FileKeys)
	Password string
	// You can add other fields here if you want...
	// Note for JSON to marshal/unmarshal, the fields need to
	// be public (start with a capital letter)
}

type EncPair struct {
	Cipher    []byte
	Signature []byte
}

// This creates a user.  It will only be called once for a user
// (unless the keystore and datastore are cleared during testing purposes)

// It should store a copy of the userdata, suitably encrypted, in the
// datastore and should store the user's public key in the keystore.

// The datastore may corrupt or completely erase the stored
// information, but nobody outside should be able to get at the stored
// User data: the name used in the datastore should not be guessable
// without also knowing the password and username.

// You are not allowed to use any global storage other than the
// keystore and the datastore functions in the userlib library.

// You can assume the user has a STRONG password
func InitUser(username string, password string) (userdataptr *User, err error) {
	if(len(username)==0){ //invalid username
			return nil, errors.New("Bad username")
	}
	var userdata User
	userdata.Username = username
	userdata.Password = password
	userdata.Pubk, userdata.Privk, err = userlib.PKEKeyGen()
	if err != nil {
		return
	}
	userdata.Signk, userdata.Vrfyk, err = userlib.DSKeyGen()
	if err != nil {
		return
	}
	userdata.Filemap = make(map[string](FileKeys))
	err = userlib.KeystoreSet(username+"Pubk", userdata.Pubk)
	if err != nil {
		return
	}
	err = userlib.KeystoreSet(username+"Vrfyk", userdata.Vrfyk)
	if err != nil {
		return
	}
	return UpdateUser(userdata)
}

func UpdateUser(userdata User) (userdataptr *User, err error) {
	password := userdata.Password
	namebyte := []byte(userdata.Username)

	uuid_root_namespace, _ := uuid.FromBytes(make([]byte, 16))
	user_UUID := uuid.NewMD5(uuid_root_namespace, namebyte)

	encpairdata, _ := AEencUser(userdata, password)

	userlib.DatastoreDelete(user_UUID)
	userlib.DatastoreSet(user_UUID, encpairdata)

	return &userdata, nil
}

func AEencUser(userdata User, password string) (encpairdata []byte, err error) {
	data, _ := json.Marshal(userdata)
	iv := userlib.RandomBytes(userlib.AESBlockSize)
	passwordbyte := []byte(password)
	passwordbyte = append(passwordbyte, passwordbyte...)
	namebyte := []byte(userdata.Username)
	if (len(passwordbyte)<=16){ //extend passwordbyte to 16 bytes
		 	if(len(passwordbyte)==0){
		 		passwordbyte = make([]byte, 16)
		 	} else {
		 		offset := 16/len(passwordbyte)
		 		for i:=0; i<offset+1; i++ {
		 			passwordbyte = append(passwordbyte, passwordbyte...)
		 		}
		 		passwordbyte = passwordbyte[:16]
		 	}
		}
	symmetricKey, _ := userlib.HMACEval(passwordbyte[:16], namebyte)
	symmetricKey = symmetricKey[:16]
	encrypteddata := userlib.SymEnc(symmetricKey, iv, data)
	signkey, _ := userlib.HMACEval(symmetricKey, namebyte)
	signkey = signkey[:16]
	signature, _ := userlib.HMACEval(signkey, data)
	var pair EncPair
	pair.Cipher = encrypteddata
	pair.Signature = signature
	encpairdata, _ = json.Marshal(pair)
	return encpairdata, nil
}

func AEdpyUser(encpairdata []byte, username string, password string) (userdata User, err error) {
	var encpair EncPair
	json.Unmarshal(encpairdata, &encpair)
	encrypteddata := encpair.Cipher
	signature := encpair.Signature
	passwordbyte := []byte(password)
	passwordbyte = append(passwordbyte, passwordbyte...)
	namebyte := []byte(username)
	if (len(passwordbyte)<=16){ //extend passwordbyte to 16 bytes
		 	if(len(passwordbyte)==0){
		 		passwordbyte = make([]byte, 16)
		 	} else {
		 		offset := 16/len(passwordbyte)
		 		for i:=0; i<offset+1; i++ {
		 			passwordbyte = append(passwordbyte, passwordbyte...)
		 		}
		 		passwordbyte = passwordbyte[:16]
		 	}
		}
	symmetricKey, _ := userlib.HMACEval(passwordbyte[:16], namebyte)
	symmetricKey = symmetricKey[:16]
	if(len(encrypteddata)<userlib.AESBlockSize){
		return userdata, errors.New("size of data to small")
	}
	data := userlib.SymDec(symmetricKey, encrypteddata)
	signkey, _ := userlib.HMACEval(symmetricKey, namebyte)
	signkey = signkey[:16]
	current_signature, _ := userlib.HMACEval(signkey, data)
	ok := userlib.HMACEqual(current_signature, signature)
	if (!ok) {
		err = errors.New("UserData corrupted")
		return userdata, err
	}
	json.Unmarshal(data, &userdata)
	return userdata, nil
}

func AEencFile(data []byte) (encpairdata []byte, symmetricKey []byte, signkey []byte) {
	iv := userlib.RandomBytes(userlib.AESBlockSize)
	symmetricKey = userlib.RandomBytes(16)
	encfiledata := userlib.SymEnc(symmetricKey, iv, data)
	signkey = userlib.RandomBytes(16)
	signature, _ := userlib.HMACEval(signkey, data)
	var pair EncPair
	pair.Cipher = encfiledata
	pair.Signature = signature
	encpairdata, _ = json.Marshal(pair)
	return
}

func AEdpyFile(file File) (filedata []byte, err error) {
	F_UUid := file.FileUUID
	encpairdata, ok := userlib.DatastoreGet(F_UUid)
	if (!ok) {
		err = errors.New("No such piece of file found in DS")
		return filedata, err
	}
	var encpair EncPair
	json.Unmarshal(encpairdata, &encpair)
	encrypteddata := encpair.Cipher
	signature := encpair.Signature
	if(len(encrypteddata)<userlib.AESBlockSize){
		return filedata, errors.New("size of data to small")
	}
	data := userlib.SymDec(file.SymmetricKey, encrypteddata)
	current_signature, _ := userlib.HMACEval(file.Signkey, data)
	ok = userlib.HMACEqual(current_signature, signature)
	if (!ok) {
		err = errors.New("FileData corrupted")
		return filedata, err
	}
	filedata = data
	return filedata, nil
}

// This fetches the user information from the Datastore.  It should
// fail with an error if the user/password is invalid, or if the user
// data was corrupted, or if the user can't be found.
func GetUser(username string, password string) (userdataptr *User, err error) {
	var userdata User
	userdataptr = &userdata
	namebyte := []byte(username)

	uuid_root_namespace, _ := uuid.FromBytes(make([]byte, 16))
	user_UUID := uuid.NewMD5(uuid_root_namespace, namebyte)

	if err != nil {
		return nil, errors.New("fuck google uuid!")
	}
	encpairdata, ok := userlib.DatastoreGet(user_UUID)
	if (!ok) {
		err = errors.New("Not found in DS")
		return nil, err
	}
	userdata, err = AEdpyUser(encpairdata, username, password)
	return userdataptr, err
}

// This stores a file in the datastore.
//
// The name of the file should NOT be revealed to the datastore!
func (userdata *User) StoreFile(filename string, data []byte) {
	if(len(filename)==0){
		return 
	}
	var currentuser User
	currentuser = *userdata
	encpairdata, symmetricKey, signkey := AEencFile(data)
	new_UUID := uuid.New()
	userlib.DatastoreSet(new_UUID, encpairdata)
	var file File
	file.Filename = filename
	file.FileUUID = new_UUID
	file.SymmetricKey = symmetricKey
	file.Signkey = signkey
	var filelist []File
	filelist = append(filelist, file)
	var newkeys FileKeys
	transferstruct := ToTransferStruct(filelist)
	transferdata, _ := json.Marshal(transferstruct)
	encpairdata, symmetricKey, signkey = AEencFile(transferdata)
	newkeys.SymmetricKey = symmetricKey
	newkeys.Signkey = signkey
	new_UUID = uuid.New() // TODO: Confirm that this logic is right.
	newkeys.FileUUID = new_UUID
	userlib.DatastoreSet(new_UUID, encpairdata)
    (currentuser.Filemap)[filename] = newkeys
	UpdateUser(currentuser)
	return
}

// This adds on to an existing file.
//
// Append should be efficient, you shouldn't rewrite or reencrypt the
// existing file, but only whatever additional information and
// metadata you need.

func (userdata *User) AppendFile(filename string, data []byte) (err error) {
	if(len(filename)==0){
		return errors.New("Bad filename")
	}
	var currentuser User
	currentuser = *userdata
	filekeys, ok := (currentuser.Filemap)[filename]
	if (!ok) {
		err = errors.New("File Not found in DS for this user")
		return err
	}
	symmetricKey := filekeys.SymmetricKey
	signkey := filekeys.Signkey
	new_UUID := filekeys.FileUUID
	transferdata, err := AEdpyTransferFile(new_UUID, symmetricKey, signkey)
	if (err != nil) {
		return err
	}
	var filelist []File
	var transferstruct FileTransfer
	json.Unmarshal(transferdata, &transferstruct)
	fileuuids := transferstruct.FileUUIDs
	symmetrickeys := transferstruct.SymmetricKeys
	signkeys := transferstruct.Signkeys
	for i := 0; i < len(fileuuids); i++ {
		var file File
		file.Filename = filename
		file.FileUUID = fileuuids[i]
		file.SymmetricKey = symmetrickeys[i]
		file.Signkey = signkeys[i]
		filelist = append(filelist, file)
	}
	var encpairdata []byte
	encpairdata, symmetricKey, signkey = AEencFile(data)
	new_UUID = uuid.New()
	userlib.DatastoreSet(new_UUID, encpairdata)
	var file File
	file.Filename = filename
	file.FileUUID = new_UUID
	file.SymmetricKey = symmetricKey
	file.Signkey = signkey
	filelist = append(filelist, file)
	transferstruct = ToTransferStruct(filelist)
	transferdata, _ = json.Marshal(transferstruct)
	iv := userlib.RandomBytes(userlib.AESBlockSize)
	encfiledata := userlib.SymEnc(filekeys.SymmetricKey, iv, transferdata)
	signature, _ := userlib.HMACEval(filekeys.Signkey, transferdata)
	var pair EncPair
	pair.Cipher = encfiledata
	pair.Signature = signature
	encpairdata, _ = json.Marshal(pair)
	userlib.DatastoreSet(filekeys.FileUUID, encpairdata)
	UpdateUser(currentuser)
	return nil
}

// This loads a file from the Datastore.
//
// It should give an error if the file is corrupted in any way.
func (userdata *User) LoadFile(filename string) (data []byte, err error) {
	if(len(filename)==0){
		return nil, errors.New("Bad filename")
	}
	var currentuser User
	currentuser = *userdata
	filekeys, ok := (currentuser.Filemap)[filename]
	if (!ok) {
		err = errors.New("File Not found in DS for this user")
		return data, err
	}
	symmetricKey := filekeys.SymmetricKey
	signkey := filekeys.Signkey
	new_UUID := filekeys.FileUUID
	transferdata, err := AEdpyTransferFile(new_UUID, symmetricKey, signkey)
	if (err != nil) {
		return data, err
	}
    var transferstruct FileTransfer
	json.Unmarshal(transferdata, &transferstruct)
	fileuuids := transferstruct.FileUUIDs
	symmetrickeys := transferstruct.SymmetricKeys
	signkeys := transferstruct.Signkeys
	for i := 0; i < len(fileuuids); i++ {
		var file File
		file.Filename = filename
		file.FileUUID = fileuuids[i]
		file.SymmetricKey = symmetrickeys[i]
		file.Signkey = signkeys[i]
		var curr_filedata []byte
		curr_filedata, err = AEdpyFile(file)
		if (err != nil) {
			return data, err
		}
		data = append(data, curr_filedata...)
	}
	return data, nil
}

// The structure definition for a file-transfer record
type FileTransfer struct {
	FileUUIDs     []uuid.UUID
	SymmetricKeys []([]byte)
	Signkeys      []([]byte)
}

func ToTransferStruct(filelist []File) (transferstruct FileTransfer) {
	for _, file := range filelist {
		transferstruct.FileUUIDs = append(transferstruct.FileUUIDs, file.FileUUID)
		transferstruct.SymmetricKeys = append(transferstruct.SymmetricKeys, file.SymmetricKey)
		transferstruct.Signkeys = append(transferstruct.Signkeys, file.Signkey)
	}
	return transferstruct
}

// You may want to define what you actually want to pass as a
// sharingRecord to serialized/deserialize in the data store.
type sharingRecord struct {
}

func AEPubencKey(recipient_PubK userlib.PKEEncKey, user_SignK userlib.DSSignKey, data []byte) (encpairdata []byte, err error) {
	encfiledata, err := userlib.PKEEnc(recipient_PubK, data)
	if (err != nil) {
		return encpairdata, err
	}
	signature, err := userlib.DSSign(user_SignK, encfiledata)
	if (err != nil) {
		return encpairdata, err
	}
	var pair EncPair
	pair.Cipher = encfiledata
	pair.Signature = signature
	encpairdata, _ = json.Marshal(pair)
	return encpairdata, nil
}

func AEPubdpyKey(recipient_PrivK userlib.PKEDecKey, user_VrfyK userlib.DSVerifyKey, encpairdata []byte) (filedata []byte, err error) {
	var encpair EncPair
	json.Unmarshal(encpairdata, &encpair)
	encrypteddata := encpair.Cipher
	signature := encpair.Signature
	err = userlib.DSVerify(user_VrfyK, encrypteddata, signature)
	if (err != nil) {
		err = errors.New("Verfication for magic failed")
		return filedata, err
	}
	filedata, err = userlib.PKEDec(recipient_PrivK, encrypteddata)
	if (err != nil) {
		return filedata, err
	}
	return filedata, nil
}

// This creates a sharing record, which is a key pointing to something
// in the datastore to share with the recipient.

// This enables the recipient to access the encrypted file as well
// for reading/appending.

// Note that neither the recipient NOR the datastore should gain any
// information about what the sender calls the file.  Only the
// recipient can access the sharing record, and only the recipient
// should be able to know the sender.

func (userdata *User) ShareFile(filename string, recipient string) (
	magic_string string, err error) {
	if(len(filename)==0){
		return magic_string, errors.New("Bad filename")
	}
	var currentuser User
	currentuser = *userdata
	recipient_PubK, ok := userlib.KeystoreGet(recipient + "Pubk")
	if (!ok) {
		err = errors.New("No such user")
		return magic_string, err
	}
	filekeys, good := (currentuser.Filemap)[filename]
	if (!good) {
		err = errors.New("File Not found in DS for this user to Share")
		return magic_string, err
	}
	var message []byte
	uuid_byte, _ := json.Marshal(filekeys.FileUUID)
	message = append(message, filekeys.SymmetricKey...)
	message = append(message, filekeys.Signkey...)
	message = append(message, uuid_byte...)
	var encpairdata []byte
	encpairdata, err = AEPubencKey(recipient_PubK, currentuser.Signk, message)
	if (err != nil) {
		return magic_string, err
	}
	magic_string = hex.EncodeToString(encpairdata)
	return magic_string, nil
}

func AEdpyTransferFile(uuid uuid.UUID, symmetricKey []byte, signkey []byte) (filedata []byte, err error) {
	encpairdata, ok := userlib.DatastoreGet(uuid)
	if (!ok) {
		err = errors.New("No such piece of transfer info found in DS")
		return filedata, err
	}
	var encpair EncPair
	json.Unmarshal(encpairdata, &encpair)
	encrypteddata := encpair.Cipher
	signature := encpair.Signature
	if(len(encrypteddata)<userlib.AESBlockSize){
		return filedata, errors.New("size of data to small")
	}
	data := userlib.SymDec(symmetricKey, encrypteddata)
	current_signature, _ := userlib.HMACEval(signkey, data)
	ok = userlib.HMACEqual(current_signature, signature)
	if (!ok) {
		err = errors.New("FileData corrupted")
		return filedata, err
	}
	filedata = data
	return filedata, nil
}

// Note recipient's filename can be different from the sender's filename.
// The recipient should not be able to discover the sender's view on
// what the filename even is!  However, the recipient must ensure that
// it is authentically from the sender.

func (userdata *User) ReceiveFile(filename string, sender string,
	magic_string string) (err error) {
	if(len(filename)==0){
		return errors.New("Bad filename to receive")
	}
	var currentuser User
	currentuser = *userdata
	encpairdata, err := hex.DecodeString(magic_string)
	if (err != nil) {
		return err
	}
	user_VrfyK, ok := userlib.KeystoreGet(sender + "Vrfyk")
	if (!ok) {
		err = errors.New("Sender is not found")
		return err
	}
	var message []byte
	message, err = AEPubdpyKey(currentuser.Privk, user_VrfyK, encpairdata)
	if (err != nil) {
		return err
	}
	symmetricKey := message[:16]
	signkey := message[16:32]
	uuid_byte := message[32:]
	var new_UUID uuid.UUID
	json.Unmarshal(uuid_byte, &new_UUID)
	var filekeys FileKeys
	filekeys.SymmetricKey = symmetricKey
	filekeys.Signkey = signkey
	filekeys.FileUUID = new_UUID
	(currentuser.Filemap)[filename] = filekeys
	UpdateUser(currentuser)
	_, err = userdata.LoadFile(filename)
	if(err!=nil){
		return err
	}
	return nil
}

// Removes access for all others.
func (userdata *User) RevokeFile(filename string) (err error) {
	var currentuser User
	currentuser = *userdata
	filekeys, ok := (currentuser.Filemap)[filename]
	if (!ok) {
		err = errors.New("File Not found in DS for this user")
		return err
	}
	symmetricKey := filekeys.SymmetricKey
	signkey := filekeys.Signkey
	new_UUID := filekeys.FileUUID
	transferdata, err := AEdpyTransferFile(new_UUID, symmetricKey, signkey)
	if (err != nil) {
		return err
	}
	var data []byte
    var transferstruct FileTransfer
	json.Unmarshal(transferdata, &transferstruct)
	fileuuids := transferstruct.FileUUIDs
	symmetrickeys := transferstruct.SymmetricKeys
	signkeys := transferstruct.Signkeys
	for i := 0; i < len(fileuuids); i++ {
		var file File
		file.Filename = filename
		file.FileUUID = fileuuids[i]
		file.SymmetricKey = symmetrickeys[i]
		file.Signkey = signkeys[i]
		var curr_filedata []byte
		curr_filedata, err = AEdpyFile(file)
		if (err != nil) {
			return err
		}
		data = append(data, curr_filedata...)
		userlib.DatastoreDelete(file.FileUUID)
	}
	userlib.DatastoreDelete(filekeys.FileUUID)
	userdata.StoreFile(filename, data)
	return nil
}

